/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.plugins.javancss;

import ch.hortis.sonar.model.Metric;
import ch.hortis.sonar.service.MeasureKey;
import org.apache.commons.lang.StringUtils;
import org.apache.maven.project.MavenProject;
import org.sonar.commons.metrics.CoreMetrics;
import org.sonar.plugins.java.AbstractJavaMavenCollector;
import org.sonar.plugins.java.JavaMeasuresRecorder;
import org.sonar.plugins.utils.CollectorUtils;
import org.sonar.plugins.utils.XmlParserException;
import org.sonar.plugins.utils.XmlReportParser;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.io.File;
import java.text.ParseException;
import java.util.HashMap;
import java.util.Map;

public class JavaNcssMavenCollector extends AbstractJavaMavenCollector {

  public JavaNcssMavenCollector(JavaMeasuresRecorder recorder) {
    super(recorder);
  }

  public boolean shouldStopOnFailure() {
    return true;
  }

  protected boolean shouldCollectIfNoTests() {
    return true;
  }

  protected boolean shouldCollectIfNoSources() {
    return false;
  }

  public void collect(MavenProject pom) {
    File xmlFile = new File(new File(pom.getBuild().getDirectory()), "javancss-raw-report.xml");
    collectMeasures(xmlFile);
  }

  protected void collectMeasures(File xmlFile) {
    try {
      XmlReportParser parser = CollectorUtils.loadAndParseXmlReport(xmlFile);
      collectPackagesCount(parser);
      collectPackageComments(parser);
      collectFileMeasures(parser);

    } catch (ParseException e) {
      throw new XmlParserException(e);

    }
  }

  private void collectPackageComments(XmlReportParser parser) throws ParseException {
    NodeList packages = parser.executeXPathNodeList("/javancss/packages/package");
    if (packages != null) {
      for (int i = 0; i < packages.getLength(); i++) {
        Element element = (Element) packages.item(i);
        String packageName = parser.getChildElementValue(element, "name");
        if (".".equals(packageName)) {
          packageName = null;
        }

        double lines = CollectorUtils.parseNumber(parser.getChildElementValue(element, "javadoc_lines")) +
            CollectorUtils.parseNumber(parser.getChildElementValue(element, "single_comment_lines")) +
            CollectorUtils.parseNumber(parser.getChildElementValue(element, "multi_comment_lines"));
        getMeasuresRecorder().createPackageMeasure(packageName, new MeasureKey(CoreMetrics.COMMENT_LINES), lines);
      }
    }
  }

  private void collectPackagesCount(XmlReportParser parser) throws ParseException {
    double packageCount = CollectorUtils.parseNumber(parser.executeXPath("count(/javancss/packages/package)"));
    getMeasuresRecorder().createProjectMeasure(new MeasureKey(CoreMetrics.PACKAGES_COUNT), packageCount);
  }

  private void collectFileMeasures(XmlReportParser parser) throws ParseException {
    NodeList classes = parser.executeXPathNodeList("/javancss/objects/object");
    if (classes != null) {
      for (int i = 0; i < classes.getLength(); i++) {
        Element element = (Element) classes.item(i);
        String fullName = parser.getChildElementValue(element, "name");
        String[] packageAndClass = parsePackageAndClass(fullName);

        createClassMeasure(packageAndClass[0], packageAndClass[1], false, CoreMetrics.NCSS, element, "ncss", parser);
        createClassMeasure(packageAndClass[0], packageAndClass[1], false, CoreMetrics.FUNCTIONS_COUNT, element, "functions", parser);
        getMeasuresRecorder().createClassMeasure(packageAndClass[0], packageAndClass[1], false, new MeasureKey(CoreMetrics.CLASSES_COUNT), 1.0);
      }
    }
    createComplexityClassMeasures(parser);

  }

  protected String[] parsePackageAndClass(String fullname) {
    String[] result = new String[2];
    result[0] = null;
    result[1] = fullname;
    if (fullname.indexOf(".") >= 0) {
      result[0] = StringUtils.substringBeforeLast(fullname, ".");
      result[1] = StringUtils.substringAfterLast(fullname, ".");
    }
    return result;
  }

  private void createClassMeasure(String packageName, String className, boolean unitTest, Metric metric, Element classElement, String key, XmlReportParser parser) throws ParseException {
    Element element = parser.getChildElement(classElement, key);
    if (element != null && element.getFirstChild() != null && element.getFirstChild().getNodeValue() != null) {
      String value = element.getFirstChild().getNodeValue();
      getMeasuresRecorder().createClassMeasure(packageName, className, unitTest, new MeasureKey(metric), CollectorUtils.parseNumber(value));
    }
  }


  private void createComplexityClassMeasures(XmlReportParser parser) throws ParseException {

    NodeList functions = parser.executeXPathNodeList("/javancss/functions/function");
    Map<String, Integer> ccnCountperClass = new HashMap<String, Integer>();
    if (functions != null) {
      for (int i = 0; i < functions.getLength(); i++) {
        Element element = (Element) functions.item(i);
        String fullFunctionName = parser.getChildElementValue(element, "name");
        Integer ccnForFunction = new Integer(parser.getChildElementValue(element, "ccn"));
        String packageAndClassName = getPackageAndClassFromFunction(fullFunctionName);
        Integer ccnSum = ccnCountperClass.get(packageAndClassName);
        if (ccnSum == null) {
          ccnSum = 0;
        }
        ccnSum += ccnForFunction;
        ccnCountperClass.put(packageAndClassName, ccnSum);
      }
    }
    for (String fullname : ccnCountperClass.keySet()) {
      Integer ccnForClass = ccnCountperClass.get(fullname);
      String[] packageAndClass = parsePackageAndClass(fullname);
      getMeasuresRecorder().createClassMeasure(packageAndClass[0], packageAndClass[1], false, new MeasureKey(CoreMetrics.COMPLEXITY_TOTAL), ccnForClass.doubleValue());
    }
  }

  protected String getPackageAndClassFromFunction
      (String
          functionName) {
    String className = StringUtils.substringBeforeLast(functionName, ".");
    return getClassName(className);
  }

  private String getClassName
      (String
          classNameCandidate) {
    int lastPointIndex = classNameCandidate.lastIndexOf(".");
    if (lastPointIndex == -1) {
      return classNameCandidate;
    }
    String packageName = StringUtils.substringBeforeLast(classNameCandidate, ".");
    String className = StringUtils.substringAfterLast(classNameCandidate, ".");
    String[] packages = StringUtils.split(packageName, '.');
    for (String pack : packages) {
      if (!isPackageName(pack)) {
        return getClassName(packageName);
      }
    }
    return packageName + "." + className;
  }

  protected boolean isPackageName
      (String
          classNameCandidate) {
    if (!Character.isLowerCase(classNameCandidate.charAt(0))) {
      return false;
    }
    for (int i = 0; i < classNameCandidate.length(); i++) {
      char currentLetter = classNameCandidate.charAt(i);
      if (!(Character.isLetterOrDigit(currentLetter) || (currentLetter == '_'))) {
        return false;
      }
    }
    return true;
  }
}
